/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001  Ludvig Strigeus
 * Copyright (C) 2001-2003 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/scumm/imuse.h,v 1.33 2003/05/23 18:35:52 jamieson630 Exp $
 *
 */

#ifndef PLAYER_V2_H
#define PLAYER_V2_H

#include "common/scummsys.h"
#include "sound/mixer.h"

struct ChannelInfo {
    uint16 time_left;
    uint16 next_cmd;
    uint16 base_freq;
    uint16 freq_delta;
    uint16 freq;
    uint16 volume;
    uint16 volume_delta;
    uint16 tempo;
    uint16 inter_note_pause;
    uint16 transpose;
    uint16 note_length;
    uint16 hull_curve;
    uint16 hull_offset;
    uint16 hull_counter;
    uint16 freqmod_table;
    uint16 freqmod_offset;
    uint16 freqmod_incr;
    uint16 freqmod_multiplier;
    uint16 freqmod_modulo;
    uint16 unknown[5];
    uint16 music_script_nr;
};

class Player_V2 {
public:
	Player_V2();
	~Player_V2();

	void startSound(int nr, byte *data);
	void stopSound(int nr);
	void stopAllSounds();
	int  getSoundStatus(int nr);

private:
	SoundMixer *_mixer;

	int sample_rate;
	int ticks_per_sample;
	int ticks_counted;
	int samples_left;
	int freq;
	int last_freq;
	int level;
	int pcjr;
	const uint16 *freqs_table;
	unsigned int decay;

	ChannelInfo channels[4];

	int   current_nr;
	byte *current_data;
	int   next_nr;
	byte *next_data;
	byte *retaddr;

	void execute_cmd(ChannelInfo *channel);
	void next_freqs(ChannelInfo *channel);
	void generate_samples(int16 *buf, int len);
	void on_timer();
	void clear_channel(int i);
	void chainNextSound();

	static void premix_proc(void *param, int16 *buf, uint len);
};

#endif
