#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/soundcard.h>

#define SAMPLE_RATE 22050
#define FREQ_HZ 236                /* Don't change! */
#define FRAGMENTS ((2<<16) | 7)
#define SAMPLE_LENGTH (SAMPLE_RATE / FREQ_HZ)

#define AUTHENTIC_SOUND
#define DECAY  0xf400              /* Depends on sample rate */
#define PCJR_DECAY  0xd000              /* Depends on sample rate */

int sound_fd;

int pc_jr = 0;

typedef unsigned short uint16;
typedef short int16;
typedef unsigned char uint8;
typedef signed char int8;

typedef struct _channel_info {
    uint16 time_left;
    uint16 next_cmd;
    uint16 base_freq;
    uint16 freq_delta;
    uint16 freq;
    uint16 volume;
    uint16 volume_delta;
    uint16 tempo;
    uint16 inter_note_pause;
    uint16 transpose;
    uint16 note_length;
    uint16 hull_curve;
    uint16 hull_offset;
    uint16 hull_counter;
    uint16 freqmod_table;
    uint16 freqmod_offset;
    uint16 freqmod_incr;
    uint16 freqmod_multiplier;
    uint16 freqmod_modulo;
    uint16 unknown[5];
    uint16 music_script_nr;
} channel_info;

void debug_channel(channel_info *channel) {
    fprintf(stdout, "chn: %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d\n",
	    channel->time_left, channel->next_cmd, 
	    channel->base_freq, channel->freq_delta, 
	    channel->freq, channel->volume, 
	    channel->volume_delta, channel->tempo);
    fprintf(stdout, "     %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d\n",
	    channel->inter_note_pause, channel->transpose,
	    channel->note_length, channel->hull_curve,
	    channel->hull_offset, channel->hull_counter,
	    channel->freqmod_table, channel->freqmod_offset);
    fprintf(stdout, "     %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d\n",
	    channel->freqmod_incr, channel->freqmod_multiplier,
	    channel->freqmod_modulo, channel->unknown[0],
	    channel->unknown[1], channel->unknown[2],
	    channel->unknown[3], channel->unknown[4]);
    fprintf(stdout, "     %5d\n",
	    channel->music_script_nr);
}


channel_info channels[8];
uint8 * retaddr;

uint8  m_lfl[128];
uint16 m_off[128];
uint8 * music_script[128];

uint8 note_lengths[] = {
    0,  
    0,  0,  2,
    0,  3,  4,
    0,  6,  8,
    0, 12, 16,
    0, 24, 32,
    0, 48, 64
};

uint16 hull_offsets[] = {
    0, 12, 24, 36, 48, 60, 
    72, 88, 104, 120, 136, 240, 
    152, 164
};

int16 hulls[] = {
    /* hull 0 */
    3, -1, 0, 0, 0, 0, 0, 0,
    0, -1, 0, 0,
    /* hull 1 (staccato) */
    3, -1, 0, 30, 0, -1, 0, 0,
    0, -1, 0, 0,
    /* hull 2 (legato) */
    3, -1, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0,
    /* hull 3 (staccatissimo) */
    3, -1, 0, 2, 0, -1, 0, 0,
    0, -1, 0, 0,
    /* hull 4 */
    3, -1, 0, 6, 0, -1, 0, 0,
    0, -1, 0, 0,
    /* hull 5 */
    3, -1, 0, 10, 0, -1, 0, 0,
    0, -1, 0, 0,
    /* hull 6 */
    60000, -1, -1000, 20, 0, 0, 0, 0,
    40000, -1, -5000,  5, 0, -1, 0, 0,
    /* hull 7 */
    50000, -1, 0, 8, 30000, -1, 0, 0,
    28000, -1, -5000,  5, 0, -1, 0, 0,
    /* hull 8 */
    60000, -1, -2000, 16, 0, 0, 0, 0,
    28000, -1, -6000,  5, 0, -1, 0, 0,
    /* hull 9 */
    55000, -1,     0,  8, 35000, -1, 0, 0,
    40000, -1, -2000, 10, 0, -1, 0, 0,
    /* hull 10 */
    60000, -1,     0,  4, -2000, 8, 0, 0,
    40000, -1, -6000,  5, 0, -1, 0, 0,
    /* hull 12 */
    0, -1,   150, 340, -150, 340, 0, -1,
    0, -1, 0, 0,
    /* hull 13  == 164*/
    20000, -1,  4000,  7, 1000, 15, 0, 0,
    35000, -1, -2000, 15, 0, -1, 0, 0,
    
    /* hull misc = 180 */
    44000, -1, -4400, 10, 0, -1, 0, 0,
    0, -1, 0, 0,

    53000, -1, -5300, 10, 0, -1, 0, 0,
    0, -1, 0, 0,

    63000, -1, -6300, 10, 0, -1, 0, 0,
    0, -1, 0, 0,

    44000, -1, -1375, 32, 0, -1, 0, 0,
    0, -1, 0, 0,

    53000, -1, -1656, 32, 0, -1, 0, 0,
    0, -1, 0, 0,

    /* hull 11 == 240 */
    63000, -1, -1968, 32, 0, -1, 0, 0,
    0, -1, 0, 0,

    44000, -1, - 733, 60, 0, -1, 0, 0,
    0, -1, 0, 0,

    53000, -1, - 883, 60, 0, -1, 0, 0,
    0, -1, 0, 0,

    63000, -1, -1050, 60, 0, -1, 0, 0,
    0, -1, 0, 0,

    44000, -1, - 488, 90, 0, -1, 0, 0,
    0, -1, 0, 0,

    53000, -1, - 588, 90, 0, -1, 0, 0,
    0, -1, 0, 0,

    63000, -1, - 700, 90, 0, -1, 0, 0,
    0, -1, 0, 0

};

uint16 freqmod_lengths[] = {
    0x1000, 0x1000, 0x20, 0x2000, 0x1000
};

uint16 freqmod_offsets[] = {
    0, 0x100, 0x200, 0x302, 0x202
};

uint8 freqmod_table[0x502] = {
    /* freqmod 1 */
    0, 3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33, 36, 39, 42, 45, 48, 51,
    54, 57, 59, 62, 65, 67, 70, 73, 75, 78, 80, 82, 85, 87, 89, 91, 94,
    96, 98, 100, 102, 103, 105, 107, 108, 110, 112, 113, 114, 116, 117,
    118, 119, 120, 121, 122, 123, 123, 124, 125, 125, 126, 126, 126, 126,
    126, 126, 126, 126, 126, 126, 126, 125, 125, 124, 123, 123, 122, 121,
    120, 119, 118, 117, 116, 114, 113, 112, 110, 108, 107, 105, 103, 102,
    100, 98, 96, 94, 91, 89, 87, 85, 82, 80, 78, 75, 73, 70, 67, 65, 62,
    59, 57, 54, 51, 48, 45, 42, 39, 36, 33, 30, 27, 24, 21, 18, 15, 12,
    9, 6, 3, 0, 253, 250, 247, 244, 241, 238, 235, 232, 229, 226, 223, 220,
    217, 214, 211, 208, 205, 202, 199, 197, 194, 191, 189, 186, 183, 181,
    178, 176, 174, 171, 169, 167, 165, 162, 160, 158, 156, 154, 153, 151,
    149, 148, 146, 144, 143, 142, 140, 139, 138, 137, 136, 135, 134, 133,
    133, 132, 131, 131, 130, 130, 130, 130, 130, 130, 130, 130, 130, 130,
    130, 131, 131, 132, 133, 133, 134, 135, 136, 137, 138, 139, 140, 142,
    143, 144, 146, 148, 149, 151, 153, 154, 156, 158, 160, 162, 165, 167,
    169, 171, 174, 176, 178, 181, 183, 186, 189, 191, 194, 197, 199, 202,
    205, 208, 211, 214, 217, 220, 223, 226, 229, 232, 235, 238, 241, 244,
    247, 250, 253,

    /*freqmod 2*/
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36,
    37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53,
    54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70,
    71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87,
    88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103,
    104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117,
    118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131,
    132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145,
    146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159,
    160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173,
    174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187,
    188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201,
    202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215,
    216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229,
    230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243,
    244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255,

    /* freqmod 3 */ 
    136, 120,
    /* freqmod 4 */ 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 
    136, 136, 136, 136, 136, 136, 136, 136,
    136, 136, 136, 136, 136, 136, 136, 136, 

    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 
    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 
    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 
    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 
    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 
    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 
    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 
    120, 120, 120, 120, 120, 120, 120, 120,
    120, 120, 120, 120, 120, 120, 120, 120, 

    41, 35, 190, 132, 225, 108, 214, 174,
    82, 144, 73, 241, 241, 187, 233, 235,
    179, 166, 219, 60, 135, 12, 62, 153,
    36, 94, 13, 28, 6, 183, 71, 222,
    179, 18, 77, 200, 67, 187, 139, 166,
    31, 3, 90, 125, 9, 56, 37, 31,
    93, 212, 203, 252, 150, 245, 69, 59,
    19, 13, 137, 10, 28, 219, 174, 50,
    32, 154, 80, 238, 64, 120, 54, 253,
    18, 73, 50, 246, 158, 125, 73, 220,
    173, 79, 20, 242, 68, 64, 102, 208,
    107, 196, 48, 183, 50, 59, 161, 34,
    246, 34, 145, 157, 225, 139, 31, 218,
    176, 202, 153, 2, 185, 114, 157, 73,
    44, 128, 126, 197, 153, 213, 233, 128,
    178, 234, 201, 204, 83, 191, 103, 214,
    191, 20, 214, 126, 45, 220, 142, 102,
    131, 239, 87, 73, 97, 255, 105, 143,
    97, 205, 209, 30, 157, 156, 22, 114,
    114, 230, 29, 240, 132, 79, 74, 119,
    2, 215, 232, 57, 44, 83, 203, 201,
    18, 30, 51, 116, 158, 12, 244, 213,
    212, 159, 212, 164, 89, 126, 53, 207,
    50, 34, 244, 204, 207, 211, 144, 45,
    72, 211, 143, 117, 230, 217, 29, 42,
    229, 192, 247, 43, 120, 129, 135, 68,
    14, 95, 80, 0, 212, 97, 141, 190,
    123, 5, 21, 7, 59, 51, 130, 31,
    24, 112, 146, 218, 100, 84, 206, 177,
    133, 62, 105, 21, 248, 70, 106, 4,
    150, 115, 14, 217, 22, 47, 103, 104,
    212, 247, 74, 74, 208, 87, 104, 118,
    250, 22, 187, 17, 173, 174, 36, 136,
    121, 254, 82, 219, 37, 67, 229, 60,
    244, 69, 211, 216, 40, 206, 11, 245,
    197, 96, 89, 61, 151, 39, 138, 89,
    118, 45, 208, 194, 201, 205, 104, 212,
    73, 106, 121, 37, 8, 97, 64, 20,
    177, 59, 106, 165, 17, 40, 193, 140,
    214, 169, 11, 135, 151, 140, 47, 241,
    21, 29, 154, 149, 193, 155, 225, 192,
    126, 233, 168, 154, 167, 134, 194, 181,
    84, 191, 154, 231, 217, 35, 209, 85,
    144, 56, 40, 209, 217, 108, 161, 102,
    94, 78, 225, 48, 156, 254, 217, 113,
    159, 226, 165, 226, 12, 155, 180, 71,
    101, 56, 42, 70, 137, 169, 130, 121,
    122, 118, 120, 194, 99, 177, 38, 223,
    218, 41, 109, 62, 98, 224, 150, 18,
    52, 191, 57, 166, 63, 137, 94, 241,
    109, 14, 227, 108, 40, 161, 30, 32,
    29, 203, 194, 3, 63, 65, 7, 132,
    15, 20, 5, 101, 27, 40, 97, 201,
    197, 231, 44, 142, 70, 54, 8, 220,
    243, 168, 141, 254, 190, 242, 235, 113,
    255, 160, 208, 59, 117, 6, 140, 126,
    135, 120, 115, 77, 208, 190, 130, 190,
    219, 194, 70, 65, 43, 140, 250, 48,
    127, 112, 240, 167, 84, 134, 50, 149,
    170, 91, 104, 19, 11, 230, 252, 245,
    202, 190, 125, 159, 137, 138, 65, 27,
    253, 184, 79, 104, 246, 114, 123, 20,
    153, 205, 211, 13, 240, 68, 58, 180,
    166, 102, 83, 51, 11, 203, 161, 16,
};

uint16 *freqs_table;
uint16  spk_freq_table[12] = {
    36484, 34436, 32503, 30679, 29007, 27332, 
    25798, 24350, 22983, 21693,  20476, 19326
};
uint16 pcjr_freq_table[12] = {
    65472, 61760, 58304, 55040, 52032, 49024, 
    46272, 43648, 41216, 38912, 36736, 34624
};

void execute_cmd(channel_info *channel) {
    uint16 value;
    int16 offset;
    uint8 *script_ptr;
    channel_info * current_channel;
    channel_info * dest_channel;

    current_channel = channel;

    if (channel->next_cmd == 0)
	return;
    script_ptr = &music_script[channel->music_script_nr][channel->next_cmd];

    for (;;) {
	uint8 opcode = *script_ptr++;
	if (opcode >= 0xf8) {
	    switch (opcode) {
	    case 0xf8: /* set hull curve */
		printf("channel[%d]: hull curve %2d\n", 
		       channel - channels, *script_ptr);
		channel->hull_curve = hull_offsets[*script_ptr/2];
		script_ptr++;
		break;

	    case 0xf9: /* set freqmod curve */
		printf("channel[%d]: freqmod curve %2d\n", 
		       channel - channels, *script_ptr);
		channel->freqmod_table = freqmod_offsets[*script_ptr/4];
		channel->freqmod_modulo = freqmod_lengths[*script_ptr/4];
		script_ptr++;
		break;

	    case 0xfd: /* clear other channel */
		value = *(int16*)script_ptr;
		printf("clear channel %d\n", value/50);
		script_ptr += 2;
		channel = &channels[value / sizeof(channel_info)];
		/* fall through */
	    case 0xfa: /* clear current channel */
		if (opcode == 0xfa)
		    printf("clear channel\n");
		channel->next_cmd   = 0;
		channel->base_freq  = 0;
		channel->freq_delta = 0;
		channel->freq = 0;
		channel->volume = 0;
		channel->volume_delta = 0;
		channel->inter_note_pause = 0;
		channel->transpose = 0;
		channel->hull_curve = 0;
		channel->hull_offset = 0;
		channel->hull_counter = 0;
		channel->freqmod_table = 0;
		channel->freqmod_offset = 0;
		channel->freqmod_incr = 0;
		channel->freqmod_multiplier = 0;
		channel->freqmod_modulo = 0;
		break;

	    case 0xfb: /* ret from subroutine */
		printf("ret from sub\n");
		script_ptr = retaddr;
		break;
	    case 0xfc: /* call subroutine */
		offset = *(int16*)script_ptr;
		printf("subroutine %d\n", offset);
		script_ptr += 2;
		retaddr = script_ptr;
		script_ptr = (void*)(int)offset;
		printf("XXX1: %p -> %04x", script_ptr, offset);
		break;
	    case 0xfe: /* loop music */
		opcode = *script_ptr++;
		offset = *(int16*)script_ptr;
		script_ptr += 2;
		printf("loop if %d to %d\n", opcode, offset);
		if (! ((uint16*) channel)[opcode/2]
		    || --((uint16*) channel)[opcode/2])
		    script_ptr += offset;
		break;

	    case 0xff: /* set parameter */
		opcode = *script_ptr++;
		value = *(uint16*)script_ptr;
		((uint16*) channel)[opcode/2] = value;
		printf("channel[%d]: set param %2d = %5d\n", 
		       channel - &channels[0], opcode, value);
		script_ptr+=2;
		if (opcode == 0)
		    goto end;
		break;
	    }
	} else { /* opcode < 0xf8 */
	    for (;;) {
		int16 note, octave;
		dest_channel = &channels[opcode >> 5];
		channel->time_left = channel->tempo * note_lengths[opcode & 0x1f];
		fprintf(stdout, "channel[%d]: @%04x note: %3d+%d len: %2d hull: %d mod: %d/%d/%d %s\n", 
			opcode>>5, script_ptr ? script_ptr - music_script[channel->music_script_nr] : 0,
			*script_ptr & 0x7f, (signed short) dest_channel->transpose, opcode & 0x1f,
			dest_channel->hull_curve, dest_channel->freqmod_table,
			dest_channel->freqmod_incr,dest_channel->freqmod_multiplier,
			*script_ptr & 0x80 ? "last":"");
		opcode = *script_ptr++;
		note = opcode & 0x7f;
		if (note != 0x7f) {
		    uint16 freq;
		    dest_channel->time_left = channel->time_left;
		    dest_channel->note_length = 
			channel->time_left - dest_channel->inter_note_pause;
		    note += dest_channel->transpose;
		    while (note < 0)
			note += 12;
		    octave = note / 12;
		    note = note % 12;
		    dest_channel->hull_offset = 0;
		    dest_channel->hull_counter = 1;
		    if (pc_jr && dest_channel == &channels[3]) {
			dest_channel->hull_curve = 180 + note * 12;
			freq = 384 - 64 * octave;
		    } else {
			freq = freqs_table[note] >> octave;
		    }
		    dest_channel->freq = dest_channel->base_freq = freq;
		}
		if ((opcode & 0x80) != 0)
		    goto end;
		opcode = *script_ptr++;
	    }
	}
    }
 end:
    channel = current_channel;
    if (channel->time_left)
	goto finish;
#if 0
    script_ptr = 0;
    if (--fc8_c)
	goto finish;

    channel->next_cmd = 0;
    sound_status[fc8_4]--;
    fc8_4 = 0;
    look_for_pending_sound();
#endif
    return;

 finish:
    channel->next_cmd = script_ptr - 
	music_script[channel->music_script_nr];
    return;
}



void next_freqs(channel_info *channel) {
    channel->volume    += channel->volume_delta;
    channel->base_freq += channel->freq_delta;
    
    channel->freqmod_offset += channel->freqmod_incr;
    if (channel->freqmod_offset > channel->freqmod_modulo)
	channel->freqmod_offset -= channel->freqmod_modulo;
    channel->freq = 
	(int) (freqmod_table[channel->freqmod_table + (channel->freqmod_offset >> 4)])
	* (int) channel->freqmod_multiplier / 256
	+ channel->base_freq;
#if 0
    printf("Freq: %d/%d, %d/%d/%d*%d %d\n",
	    channel->base_freq, (int16)channel->freq_delta,
	    channel->freqmod_table, channel->freqmod_offset,
	    channel->freqmod_incr, channel->freqmod_multiplier,
	    channel->freq);
#endif
    
    if (channel->note_length && !--channel->note_length) {
	channel->hull_offset += 16;
	channel->hull_counter = 1;
    }

    if (!--channel->time_left) {
	execute_cmd(channel);
    }
#if 0
    fprintf(stdout, "channel[%d]: freq %d hull %d/%d/%d\n", 
	    channel - &channels[0], channel->freq, channel->hull_curve, channel->hull_offset, channel->hull_counter);
#endif
    
    if (channel->hull_counter && !--channel->hull_counter) {
	for (;;) {
	    int16 *hull_ptr = hulls
		+ channel->hull_curve + channel->hull_offset/2;
	    if (hull_ptr[1] == -1) {
		channel->volume = hull_ptr[0];
		if (hull_ptr[0] == 0)
		    channel->volume_delta = 0;
		channel->hull_offset += 4;
	    } else {
		channel->volume_delta = hull_ptr[0];
		channel->hull_counter = hull_ptr[1];
		channel->hull_offset += 4;
		break;
	    }
	}
    }
}

#ifdef AUTHENTIC_SOUND
/**
 * This simulates the pc speaker sound, which is driven
 * by the 8253 (square wave generator) and a low-band filter.
 */
void produce_spk_sound() {
    int8 sample[SAMPLE_LENGTH];
    static int sample_offset = 0;
    static int last_freq=100;
    int winning_channel = -1;
    int i, j;
    int freq;
    static uint level = 0;
    for (j = 0; j < SAMPLE_LENGTH; j++) {
	sample[j] = 0;
    }
    for (i = 0; i < 4; i++) {
	if (!channels[i].time_left)
	    continue;
	next_freqs(&channels[i]);
	if (winning_channel == -1
	    && channels[i].volume
	    && channels[i].time_left) {
	    winning_channel = i;
	}
    }
    if (winning_channel != -1) {
	freq = channels[winning_channel].freq;
    } else {
	freq = 0;
    }

    for (j = 0; j < SAMPLE_LENGTH; j++) {
	level = (level * DECAY) >> 16;
	if (sample_offset < freq*500)
	    level += 0xffff-DECAY;
	sample[j] = (level >> 8);
	
	sample_offset += 1193000*1000 / SAMPLE_RATE;
	if (sample_offset >= last_freq*1000) {
	    sample_offset -= last_freq*1000;
	    last_freq = freq;
	}
    }

#if 0
    fprintf(stdout, "channel[%d]: freq %d  ; %d\n", winning_channel, freq, sample_offset);
#endif
    fflush(stdout);
    for (j = 0; j < SAMPLE_LENGTH;)
	j += write(sound_fd, sample + j, SAMPLE_LENGTH - j);
}

#else
void produce_spk_sound() {
    int8 sample[SAMPLE_LENGTH];
    static int sample_offset = 0;
    int winning_channel = -1;
    int i, j;
    float freq, volume;
    for (j = 0; j < SAMPLE_LENGTH; j++) {
	sample[j] = 0;
    }
    for (i = 0; i < 4; i++) {
	if (!channels[i].time_left)
	    continue;
	next_freqs(&channels[i]);
	if (winning_channel == -1
	    && channels[i].volume
	    && channels[i].time_left) {
	    winning_channel = i;
	}
    }
    if (winning_channel != -1) {
	i = winning_channel;
	freq = channels[i].freq;
	volume = channels[i].volume;
	volume /= 3.0;
	freq = 1193000 / freq;


	for (j = 0; j < SAMPLE_LENGTH; j++) {
	    sample[j] += (int8) 
		(volume * 100
#ifdef SAEGEZAHN
		 * (sample_offset < SAMPLE_RATE/2
		    ? sample_offset - SAMPLE_RATE/4
		    : 3*SAMPLE_RATE/4 - sample_offset) /(SAMPLE_RATE/4)
#elif defined(RECHTECK)
		 * (sample_offset < SAMPLE_RATE/2 ? 1 : -1)
#else
		 * sin((double)sample_offset*2*3.145926 / SAMPLE_RATE)
#endif
		 );
	    sample_offset += freq;
	    sample_offset %= SAMPLE_RATE;
	}

#if 0
	fprintf(stdout, "channel[%d]: freq %d = %0f  ; volume %f  %d\n", i, channels[i].freq, freq, volume, sample_offset);
#endif
    }
    fflush(stdout);
    for (j = 0; j < SAMPLE_LENGTH;)
	j += write(sound_fd, sample + j, SAMPLE_LENGTH - j);
}
#endif

#define STEP 0x10000
#define UPDATESTEP ((STEP * SAMPLE_RATE) / (111860*2))
#define MAX_OUTPUT 0x7fff

#define FB_WNOISE 0x12000       /* bit15.d(16bits) = bit0(out) ^ bit2 */
#define FB_PNOISE 0x08000       /* JH 981127 - fixes Do Run Run */
#define NG_PRESET 0x0f35

int SN76496_count[4];
int SN76496_output[4];
int SN76496_voltable[16];

void init_pcjr()
{
  int i;
  double out;
  int gain;
  /* increase max output basing on gain (0.2 dB per step) */
  out = MAX_OUTPUT / 3;
  gain = 10;
  while (gain-- > 0)
    out *= 1.023292992;         /* = (10 ^ (0.2/20)) */
  /* build volume table (2dB per step) */
  for (i = 0; i < 15; i++) {
    /* limit volume to avoid clipping */
    if (out > MAX_OUTPUT / 3)
      SN76496_voltable[i] = MAX_OUTPUT / 3;
    else
      SN76496_voltable[i] = out;
    printf("%2d: %4x\n", i, SN76496_voltable[i]);
    out /= 1.258925412;         /* = 10 ^ (2/20) = 2dB */
  }
  SN76496_voltable[15] = 0;
}

void produce_pcjr_sound() {
    uint16 sample[SAMPLE_LENGTH];
    int i, j;
    uint16 freq, vol;
    static int16 lastvol = 0;
    memset(sample, 0, sizeof(sample));
    for (i = 0; i < 4; i++) {
	if (!channels[i].time_left)
	    continue;
	next_freqs(&channels[i]);

	freq = channels[i].freq >> 6;
	vol  = (65535 - channels[i].volume) >> 12;
	if (!channels[i].volume
	    && !channels[i].time_left) {
	  if (SN76496_count[i] >= SAMPLE_LENGTH * STEP)
	    SN76496_count[i] -= STEP * SAMPLE_LENGTH;
	} else if (i < 3) {
	  int period = UPDATESTEP * freq;
	  if (period == 0)
	    period = UPDATESTEP;

	  for (j = 0; j < SAMPLE_LENGTH; j++) {
	    unsigned int volume = 0;
	    
	    if (SN76496_output[i])
	      volume += SN76496_count[i];
	    
	    SN76496_count[i] -= STEP;
	    while (SN76496_count[i] <= 0) {
	      SN76496_count[i] += period;
	      if (SN76496_count[i] > 0) {
		SN76496_output[i] ^= 1;
		if (SN76496_output[i])
		  volume += period;
		break;
	      }
	      SN76496_count[i] += period;
	      volume += period;
	    }
	    
	    if (SN76496_output[i])
	      volume -= SN76496_count[i];
	    volume *= SN76496_voltable[vol];
	    sample[j] += volume / STEP;
	  }
	} else {
	  int left, period, RNG, NoiseFB;

	  int n = (freq & 3);
	  if (n == 3) {
	    period = 2 * UPDATESTEP * (channels[2].freq>>6);
	    if (!period)
	      period = 2 * UPDATESTEP;
	  } else
	    period = UPDATESTEP << (5 + n);

	  RNG = NG_PRESET;
	  NoiseFB = (freq & 4) ? FB_WNOISE : FB_PNOISE;
	  SN76496_output[3] = RNG & 1;

	  for (j = 0; j < SAMPLE_LENGTH; j++) {
	    unsigned int volume = 0;
	      
	    if (SN76496_output[3])
	      volume += SN76496_count[3];
	    SN76496_count[3] -= STEP;
	    while (SN76496_count[3] <= 0) {
	      if (RNG & 1)
		RNG ^= NoiseFB;
	      RNG >>= 1;
	      SN76496_output[3] = RNG & 1;
	      SN76496_count[3] += period;
	      if (SN76496_output[3])
		volume += period;
	    }
	    if (SN76496_output[3])
	      volume -= SN76496_count[3];
	    
	    volume *= SN76496_voltable[vol];
	    sample[j] += volume / STEP;
	  }
	}
#if 0
	fprintf(stdout, "channel[%d]: freq %d %.1f ; volume %d \n", i, freq, 111860.0/freq,  vol);
#endif
    }
    for (j = 0; j < SAMPLE_LENGTH; j++) {
	lastvol = ((int)lastvol * PCJR_DECAY + (int)sample[j] * (0x10000-PCJR_DECAY)) >> 16;
	sample[j] = lastvol;
    }

    fflush(stdout);
    for (j = 0; j < 2*SAMPLE_LENGTH;)
	j += write(sound_fd, sample + j, 2*SAMPLE_LENGTH - j);
#if 0
    for (j = 0; j < 2*SAMPLE_LENGTH;)
	j += write(3, sample + j, 2*SAMPLE_LENGTH - j);
#endif
}


void init_channel(channel_info *channel) {
    channel->next_cmd   = 0;
    channel->base_freq  = 0;
    channel->freq_delta = 0;
    channel->freq = 0;
    channel->volume = 0;
    channel->volume_delta = 0;
    channel->inter_note_pause = 0;
    channel->transpose = 0;
    channel->hull_curve = 0;
    channel->hull_offset = 0;
    channel->hull_counter = 0;
    channel->freqmod_table = 0;
    channel->freqmod_offset = 0;
    channel->freqmod_incr = 0;
    channel->freqmod_multiplier = 0;
    channel->freqmod_modulo = 0;
    channel->unknown[0] = 0;
    channel->unknown[1] = 0;
    channel->unknown[2] = 0;
    channel->unknown[3] = 0;
    channel->unknown[4] = 0;
    channel->tempo = 0;
    channel->note_length = 0;
}


void load_music(int nr) {
    int i;
    uint16 *music = (uint16*) music_script[nr];
    int offset = pc_jr ? 7 : 3;
    for (i = 0; i< 4; i++) {
	init_channel(&channels[i]);
	if (music[offset+i]) {
	    channels[i].music_script_nr = nr;
	    channels[i].next_cmd = music[offset+i];
	    channels[i].time_left = 1;
	}
    }

    while (channels[0].time_left
	   || channels[1].time_left
	   || channels[2].time_left
	   || channels[3].time_left) {
	if (pc_jr)
	    produce_pcjr_sound();
	else
	    produce_spk_sound();
    }
}

void read_level0() {
    FILE * file;
    uint8  b;
    uint16 w;
    int i;
    file = fopen("00.lfl", "rb");

    fread(&w, 2, 1, file);
    w ^= 0xffff;

    /* skip object */
    fread(&w, 2, 1, file);
    w ^= 0xffff;
    fseek(file, w, SEEK_CUR);

    /* skip unknowns */
    fread(&b, 1, 1, file);
    b ^= 0xff;
    fseek(file, 3*b, SEEK_CUR);

    /* skip costumes */
    fread(&b, 1, 1, file);
    b ^= 0xff;
    fseek(file, 3*b, SEEK_CUR);

    /* skip scripts */
    fread(&b, 1, 1, file);
    b ^= 0xff;
    fseek(file, 3*b, SEEK_CUR);

    /* read music */
    fread(&b, 1, 1, file);
    b ^= 0xff;
    fread(&m_lfl, 1, b, file);
    fread(&m_off, 2, b, file);
    for (i = 0; i < b; i++) {
	m_lfl[i] ^= 0xff;
	m_off[i] ^= 0xffff;
    }
    fclose(file); 
}

void read_music(int nr) {
    char lvl[20];
    FILE * file;
    int i;
    uint16 len;
    sprintf(lvl, "%02d.lfl", m_lfl[nr]);
    printf("Opening %s\n", lvl);
    file = fopen(lvl, "rb");
    fseek(file, m_off[nr], SEEK_CUR);
    fread(&len, 2, 1, file);
    len ^= 0xffff;
    music_script[nr] = malloc(len);
    fseek(file, -2, SEEK_CUR);
    fread(music_script[nr], 1, len, file);
    for (i = 0; i < len; i++)
	music_script[nr][i] ^= 0xff;
    fclose(file);
}

void open_sound() {
    int param;
#ifndef WRITE_TO_RAW_FILE
    sound_fd = open("/dev/dsp", O_WRONLY);

    param = FRAGMENTS;
    ioctl(sound_fd, SNDCTL_DSP_SETFRAGMENT, &param);
    printf ("Frags %d, ", param);
    param = pc_jr ? AFMT_S16_LE : AFMT_U8;
    ioctl(sound_fd, SNDCTL_DSP_SETFMT, &param);
    printf ("Format %d, ", param);
    param = 1;
    ioctl(sound_fd, SNDCTL_DSP_CHANNELS, &param);
    printf ("channels %d, ", param);

    param = SAMPLE_RATE;
    ioctl(sound_fd, SNDCTL_DSP_SPEED, &param);
    printf ("Sample rate: %d\n", param);
#else
    sound_fd = open("intro.raw", O_CREAT|O_WRONLY, 0666);
#endif
}

int main(int argc, char** argv) {
    int nr, i;
    if (argc > 2)
      pc_jr = 1;
    open_sound();
    read_level0();
    if (argc > 1)
	nr = atoi(argv[1]);
    else
	nr = 95;
    srand(1234);
    /* freqmod_table[302-502] contains random values.  This code replaces
       this with different ones. Try it and say whether you hear the difference.
    for (i = 0x302; i < 0x502; i++)
	freqmod_table[i] = (rand()>> 8) & 0xff;
    */
    read_music(nr);
    if (pc_jr)
      init_pcjr();

    freqs_table = pc_jr ? pcjr_freq_table : spk_freq_table;

    load_music(nr);
    return 0;
}

